'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var ramda = require('ramda');

// import { getHash } from './hashing';
/** @internal */
const createPath = (prevMap, final) => {
    const path = [final];
    let prev = prevMap.get(ramda.toString(final));
    while (prev != null) {
        path.unshift(prev);
        prev = prevMap.get(ramda.toString(prev));
    }
    return path;
};

const TOP = 0;
/* eslint-disable no-bitwise */
// get parent index
const parent = (i) => ((i + 1) >>> 1) - 1;
// double + 1
const left = (i) => (i << 1) + 1;
// double + 2
const right = (i) => (i + 1) << 1;
/* eslint-enable no-bitwise */
class PriorityQueue {
    comparator;
    heap = [];
    constructor(comparator) {
        this.comparator = comparator;
    }
    size() {
        return this.heap.length;
    }
    isEmpty() {
        return this.size() === 0;
    }
    peek() {
        return this.heap[TOP];
    }
    replace(value) {
        const replacedValue = this.peek();
        this.heap[TOP] = value;
        this.siftDown();
        return replacedValue;
    }
    push(value) {
        this.heap.push(value);
        this.siftUp();
    }
    pop() {
        const poppedValue = this.peek();
        const bottom = this.size() - 1;
        if (bottom > TOP) {
            this.swap(TOP, bottom);
        }
        this.heap.pop();
        this.siftDown();
        return poppedValue;
    }
    greater(i, j) {
        return this.comparator(this.heap[i], this.heap[j]);
    }
    swap = (i, j) => {
        const a = this.heap[i];
        const b = this.heap[j];
        this.heap[j] = a;
        this.heap[i] = b;
    };
    siftUp() {
        let node = this.size() - 1;
        while (node > TOP && this.greater(node, parent(node))) {
            this.swap(node, parent(node));
            node = parent(node);
        }
    }
    siftDown() {
        let node = TOP;
        while ((left(node) < this.size() && this.greater(left(node), node)) ||
            (right(node) < this.size() && this.greater(right(node), node))) {
            const maxChild = right(node) < this.size() && this.greater(right(node), left(node)) ? right(node) : left(node);
            this.swap(node, maxChild);
            node = maxChild;
        }
    }
}
const priorityQueue = (comparator) => {
    return new PriorityQueue(comparator);
};

const aStarAssocTraversal = function* (getNextStates, estimateRemainingCost, initial) {
    const cameFrom = new Map();
    const initialS = ramda.toString(initial);
    const gScore = new Map([[initialS, 0]]);
    const fScore = new Map([[initialS, estimateRemainingCost(initial)]]);
    const queue = priorityQueue((a, b) => {
        const aScore = fScore.get(ramda.toString(a));
        const bScore = fScore.get(ramda.toString(b));
        return aScore < bScore;
    });
    queue.push(initial);
    while (!queue.isEmpty()) {
        const current = queue.pop();
        const currentS = ramda.toString(current);
        yield [gScore.get(currentS), createPath(cameFrom, current)];
        const nextStates = getNextStates(current);
        for (const [nextState, cost] of nextStates) {
            const nextStateS = ramda.toString(nextState);
            const tentativeGScore = gScore.get(currentS) + cost;
            if (tentativeGScore < (gScore.get(nextStateS) ?? Infinity)) {
                cameFrom.set(nextStateS, current);
                gScore.set(nextStateS, tentativeGScore);
                fScore.set(nextStateS, tentativeGScore + estimateRemainingCost(nextState));
                queue.push(nextState);
            }
        }
    }
    return undefined;
};
const aStarTraversal = function* (getNextStates, getCost, estimateRemainingCost, initial) {
    const nextAssoc = (state) => getNextStates(state).map(n => [n, getCost(state, n)]);
    yield* aStarAssocTraversal(nextAssoc, estimateRemainingCost, initial);
};
/**
 * Performs a best-first search
 * using the A* search algorithm, starting with the state @initial@, generating
 * neighboring states and their associated costs with @next@, and an estimate of
 * the remaining cost with @remaining@. This returns a path to a state for which
 * @found@ returns 'True'. If @remaining@ is strictly a lower bound on the
 * remaining cost to reach a solved state, then the returned path is the
 * shortest path. Returns 'Nothing' if no path to a solved state is possible.
 *
 * @public
 * @param getNextStates - Function to generate list of neighboring states with associated transition costs given the current state
 * @param estimateRemainingCost - Estimate on remaining cost given a state
 * @param determineIfFound - Predicate to determine if solution found. `aStar` returns the shortest path to the first state for which this predicate returns `true`
 * @param start - starting state
 * @returns [Total cost, list of steps] for the first path found which satisfies the given predicate
 */
const aStarAssoc = (getNextStates, estimateRemainingCost, determineIfFound, start) => {
    const visited = [];
    for (const [value, pathTo] of aStarAssocTraversal(getNextStates, estimateRemainingCost, start)) {
        const current = pathTo[pathTo.length - 1];
        visited.push(current);
        if (determineIfFound(current))
            return [value, pathTo, visited];
    }
    return undefined;
};
/**
 * Performs a best-first search
 * using the A* search algorithm, starting with the state @initial@, generating
 * neighboring states with `next`, their cost with @cost@, and an estimate of
 * the remaining cost with `remaining`. This returns a path to a state for which
 * `found` returns `true`. If `remaining` is strictly a lower bound on the
 * remaining cost to reach a solved state, then the returned path is the
 * shortest path. Returns `undefined` if no path to a solved state is possible.
 *
 * @public
 * @param getNextStates - Function to generate list of neighboring states given the current state
 * @param getCost - Function to generate transition costs between neighboring states
 * @param estimateRemainingCost - Estimate on remaining cost given a state
 * @param determineIfFound - Predicate to determine if solution found. `aStar` returns the shortest path to the first state for which this predicate returns `true`
 * @param initial - Initial state
 * @returns - [Total cost, list of steps] for the first path found which satisfies the given predicate
 */
const aStar = (getNextStates, getCost, estimateRemainingCost, determineIfFound, initial) => {
    const visited = [];
    for (const [value, pathTo] of aStarTraversal(getNextStates, getCost, estimateRemainingCost, initial)) {
        const current = pathTo[pathTo.length - 1];
        visited.push(current);
        if (determineIfFound(current))
            return [value, pathTo, visited];
    }
    return undefined;
};

/**
 * Performs a breadth-first traversal over a set of states.
 * Starting with `initial`, and generating neighboring states with `next`.
 * This generator yields each state as it is visited.
 * Caution: If your states grow infinitely, so will this generator
 *
 * @param next
 * @param start
 */
const breadthFirstTraversal = function* (next, start) {
    const visited = new Set();
    // we queue a pair of values and the path through to get there
    const queue = [[start, []]];
    while (queue.length) {
        const [value, pathSoFar] = queue.shift();
        const asStr = ramda.toString(value);
        if (visited.has(asStr))
            continue;
        yield [value, [...pathSoFar, value]];
        visited.add(asStr);
        const nextPathSoFar = [...pathSoFar, value];
        queue.push(...next(value)
            .filter(v => !visited.has(ramda.toString(v)))
            .map(v => [v, nextPathSoFar]));
    }
};
/**
 * Performs a breadth-first-search (bfs) over a set of states.
 * Starting with `initial`, and generating neighboring states with `next`
 * Returns a path to a state when `found` returns `true`
 * Returns `undefined` if no path is possible.
 *
 * @public
 * @param next - Function to generate "next" states given a current state
 * @param found - Predicate to determine if solution found. `bfs` returns a path to the first state for which this predicate returns `true`.
 * @param start - Initial state
 * @returns First path found to a state matching the predicate, `undefined` if no such path exists.
 */
const breadthFirstSearch = (next, found, start) => {
    const visited = [];
    for (const [value, pathTo] of breadthFirstTraversal(next, start)) {
        visited.push(value);
        if (found(value))
            return [value, pathTo, visited];
    }
    return undefined;
};

/**
 * Performs a depth-first traversal over a set of states.
 * Starting with `initial`, and generating neighboring states with `next`.
 * This generator yields each state as it is visited.
 * Caution: If your states grow infinitely, so will this generator
 *
 * @param next
 * @param initial
 */
const depthFirstTraversal = function* (next, start) {
    const visited = new Set();
    // we stack a pair of values and the path through to get there
    const stack = [[start, []]];
    while (stack.length) {
        const [value, pathSoFar] = stack.pop();
        const asStr = ramda.toString(value);
        if (visited.has(asStr))
            continue;
        yield [value, [...pathSoFar, value]];
        visited.add(asStr);
        const nextPathSoFar = [...pathSoFar, value];
        stack.push(...next(value)
            .filter(v => !visited.has(ramda.toString(v)))
            .map(v => [v, nextPathSoFar])
            .reverse());
    }
};
/**
 * Performs a depth-first search over a set
 * of states, starting with @initial@ and generating neighboring states with
 * @next@. It returns a depth-first path to a state for which @found@ returns
 * 'True'. Returns 'Nothing' if no path is possible.
 *
 * @public
 * @param next - Function to generate "next" states given a current state. These should be given in the order in which states should be pushed onto the stack, i.e. the "last" state in the Foldable will be the first one visited.
 * @param found - Predicate to determine if solution found. `dfs` returns a path to the first state for which this predicate returns `true`.
 * @param initial - Initial state
 * @returns First path found to a state matching the predicate, or `undefined` if no such path exists.
 */
const depthFirstSearch = (next, found, start) => {
    const visited = [];
    for (const [value, pathTo] of depthFirstTraversal(next, start)) {
        visited.push(value);
        if (found(value))
            return [value, pathTo, visited];
    }
    return undefined;
};

const dijkstraAssocTraversal = function* (getNextStates, initial) {
    const prevMap = new Map();
    const costMap = new Map([[ramda.toString(initial), 0]]);
    const queue = priorityQueue((a, b) => {
        const aCost = costMap.get(ramda.toString(a)) ?? Infinity;
        const bCost = costMap.get(ramda.toString(b)) ?? Infinity;
        return aCost < bCost;
    });
    queue.push(initial);
    while (!queue.isEmpty()) {
        const current = queue.pop();
        const currentS = ramda.toString(current);
        yield [costMap.get(currentS), createPath(prevMap, current)];
        const visitCost = costMap.get(currentS) ?? Infinity;
        const nextStates = getNextStates(current);
        for (const [nextState, nextCost] of nextStates) {
            const nextStateS = ramda.toString(nextState);
            const altCost = visitCost + nextCost;
            if (altCost < (costMap.get(nextStateS) ?? Infinity)) {
                costMap.set(nextStateS, altCost);
                prevMap.set(nextStateS, current);
                queue.push(nextState);
            }
        }
    }
    return undefined;
};
const dijkstraTraversal = function* (getNextStates, getCost, initial) {
    const nextAssoc = (state) => getNextStates(state).map(n => [n, getCost(state, n)]);
    yield* dijkstraAssocTraversal(nextAssoc, initial);
};
/**
 * This API to Dijkstra's algorithm is useful in the common case when next
 * states and their associated transition costs are generated together.
 *
 * Dijkstra's algorithm can be viewed as a generalized search, with the search
 * container being a heap, with the states being compared without regard to
 * cost, with the shorter paths taking precedence over longer ones, and with
 * the stored state being (cost so far, state).
 * This implementation makes that transformation, then transforms that result
 * back into the desired result from `dijkstraAssoc`
 *
 * @public
 * @param getNextStates - function to generate list of neighboring states with associated transition costs given the current state
 * @param determineIfFound - Predicate to determine if solution found. 'dijkstraAssoc' returns the shortest path to the first state for which this predicate returns `true`
 * @param initial - Initial state
 * @returns [Total cost, list of steps] for the first path found which satisfies the given predicate
 */
const dijkstraAssoc = (getNextStates, determineIfFound, initial) => {
    const visited = new Map();
    for (const [value, pathTo] of dijkstraAssocTraversal(getNextStates, initial)) {
        const current = pathTo[pathTo.length - 1];
        visited.set(ramda.toString(current), current);
        if (determineIfFound(current))
            return [value, pathTo, [...visited.values()]];
    }
    return [undefined, undefined, [...visited.values()]];
};
/**
 * Performs a shortest-path search over
 * a set of states using Dijkstra's algorithm, starting with `initial`,
 * generating neighboring states with `next`, and their incremental costs with
 * `costs`. This will find the least-costly path from an initial state to a
 * state for which `found` returns `true`. Returns `null` if no path to a
 * solved state is possible.
 *
 * @public
 * @param getNextStates - Function to generate list of neighboring states given the current state
 * @param getCost - Function to generate transition costs between neighboring states
 * @param determineIfFound - Predicate to determine if solution found. 'dijkstra' returns the shortest path to the first state for which this predicate returns `true`
 * @param initial
 * @returns
 */
const dijkstra = (getNextStates, getCost, determineIfFound, initial) => {
    const nextAssoc = (state) => getNextStates(state).map(n => [n, getCost(state, n)]);
    return dijkstraAssoc(nextAssoc, determineIfFound, initial);
};

exports.aStar = aStar;
exports.aStarAssoc = aStarAssoc;
exports.aStarAssocTraversal = aStarAssocTraversal;
exports.aStarTraversal = aStarTraversal;
exports.breadthFirstSearch = breadthFirstSearch;
exports.breadthFirstTraversal = breadthFirstTraversal;
exports.depthFirstSearch = depthFirstSearch;
exports.depthFirstTraversal = depthFirstTraversal;
exports.dijkstra = dijkstra;
exports.dijkstraAssoc = dijkstraAssoc;
exports.dijkstraAssocTraversal = dijkstraAssocTraversal;
exports.dijkstraTraversal = dijkstraTraversal;
//# sourceMappingURL=index.cjs.map
